#include <sys/socket.h> /* for socket(), bind(), getsockname() etc. */
#include <arpa/inet.h>  /* for inet_aton()                          */
#include <stdio.h>      /* for perror()                             */
#include <string.h>     /* for memset()                             */
#include <unistd.h>     /* for sleep() and write()                  */
#include <errno.h>      /* for EINPROGRESS                          */
#include <fcntl.h>      /* for fcntl()                              */

#include "socket.hh"

void Socket::BecomeListenerTCP(int port)
{
    int sock, RetryCount=0;
    
RetrySocket:
    sock = socket(AF_INET, SOCK_STREAM, 0);
    if(sock < 0)
    {
        perror("socket");
        /* FIXME: Now what? */
        sleep(1);
        if(++RetryCount<50)goto RetrySocket;
    }

    struct sockaddr_in mysock;

    memset(&mysock, 0, sizeof(mysock));

    mysock.sin_family = AF_INET;
    mysock.sin_addr.s_addr = htonl(INADDR_ANY);
    /* FIXME: Allow this to be selected^ */
    mysock.sin_port = htons(port);
    
    int dummy = 1;
    setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &dummy, sizeof dummy);
    
RetryBind:
    if(bind(sock, (const struct sockaddr *)&mysock, sizeof(mysock)) < 0)
    {
        perror("bind");
        /* FIXME: Now what? */
        sleep(1);
        if(++RetryCount<50)goto RetryBind;
    }
    
    listen(sock, 5);
    
    Type          = Socket::SockTypeListening;
    fd            = sock;
    Peer.Protocol = SockProtocolTCP;
    Peer.DestPort = port;
}

void Socket::BecomeListenerUDP(int port)
{
    int sock, RetryCount=0;
    
RetrySocket:
    sock = socket(AF_INET, SOCK_DGRAM, 0);
    if(sock < 0)
    {
        perror("socket");
        /* FIXME: Now what? */
        sleep(1);
        if(++RetryCount<50)goto RetrySocket;
    }

    struct sockaddr_in mysock;
    memset(&mysock, 0, sizeof(mysock));

    mysock.sin_family = AF_INET;
    mysock.sin_addr.s_addr = htonl(INADDR_ANY);
    /* FIXME: Allow this to be selected^ */
    mysock.sin_port = htons(port);
    
    int dummy = 1;
    setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &dummy, sizeof dummy);
    
RetryBind:
    if(bind(sock, (const struct sockaddr *)&mysock, sizeof(mysock)) < 0)
    {
        perror("bind");
        /* FIXME: Now what? */
        sleep(1);
        if(++RetryCount<50)goto RetryBind;
    }
    
    Type          = Socket::SockTypeListening;
    fd            = sock;
    Peer.Protocol = SockProtocolUDP;
    Peer.DestPort = port;
}

void Socket::BecomeLocalConnecterTCP(int port, int srcport)
{
    BecomeConnecterTCP(port, "127.0.0.1", srcport);
}

void Socket::BecomeConnecterTCP(int port, const char *ip, int srcport)
{
    int sock, RetryCount=0;
    
RetrySocket:
    sock = socket(AF_INET, SOCK_STREAM, 0);
    if(sock < 0)
    {
        perror("socket");
        /* FIXME: Now what? */
        sleep(1);
        if(++RetryCount<50)goto RetrySocket;
    }
    
    Peer.AddrLen = sizeof(Peer.Addr);
    memset(&Peer.Addr, 0, sizeof(Peer.Addr));
    Peer.Addr.sin_family = AF_INET;
    
    if(srcport != 0)
    {
        Peer.Addr.sin_addr.s_addr = htonl(INADDR_ANY);
        /* FIXME: Allow this to be selected^ */
        Peer.Addr.sin_port = htons(srcport);
        
        int dummy = 1;
        setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &dummy, sizeof dummy);
        
RetryBind:
        if(bind(sock, (const struct sockaddr *)&Peer.Addr, Peer.AddrLen) < 0)
        {
            perror("bind");
            /* FIXME: Now what? */
            sleep(1);
            if(++RetryCount<50)goto RetryBind;
        }
    }

    Peer.Addr.sin_port = htons(port);
    inet_aton(ip, &Peer.Addr.sin_addr);
    
RetryConnect:
    fcntl(sock, F_SETFL, fcntl(sock, F_GETFL) | O_NONBLOCK);
    int res = connect(sock, (const struct sockaddr *)&Peer.Addr, Peer.AddrLen);
    if(res < 0 && errno != EINPROGRESS)
    {
        perror("connect");
        /* FIXME: Now what? */
        sleep(1);
        if(++RetryCount<50)goto RetryConnect;
    }
    fcntl(sock, F_SETFL, fcntl(sock, F_GETFL) & ~O_NONBLOCK);
    
    if(getsockname(sock, (struct sockaddr *)&Peer.Addr, &Peer.AddrLen) < 0)
    {
        perror("getsockname");
        /* FIXME: Now what? */
    }
    fprintf(stderr, "Local TCP port is %u\n", ntohs(Peer.Addr.sin_port));

    Type          = SockTypeConnecting;
    fd            = sock;
    Peer.Protocol = SockProtocolTCP;
    Peer.DestPort = port;
    
    if(!res)
        Type = SockTypeConnected;
}

void Socket::BecomeLocalSenderUDP(int port)
{
    /* Note: port is the destination port, not the source port */
    
    int sock, RetryCount=0;
    
RetrySocket:
    sock = socket(AF_INET, SOCK_DGRAM, 0);
    if(sock < 0)
    {
        perror("socket");
        /* FIXME: Now what? */
        sleep(1);
        if(++RetryCount<50)goto RetrySocket;
    }
    
    Peer.AddrLen = sizeof(Peer.Addr);
    if(getsockname(sock, (struct sockaddr *)&Peer.Addr, &Peer.AddrLen) < 0)
    {
        perror("getsockname");
        /* FIXME: Now what? */
    }
    
    fprintf(stderr, "Local UDP port is %u\n", ntohs(Peer.Addr.sin_port));

    Type          = Socket::SockTypeConnected;
    fd            = sock;
    Peer.Protocol = SockProtocolUDP;
    Peer.DestPort = port;
    /* source port is Peer.Addr.sin_port */
}

DataPacket Socket::AcceptTCP(const Socket &parent)
{
    int RetryCount = 0;
    
    Peer = parent.Peer;
    
RetryAccept:
    Peer.AddrLen = sizeof(Peer.Addr);

    int sock = accept(parent.fd, (struct sockaddr *)&Peer.Addr, &Peer.AddrLen);
    if(sock < 0)
    {
        perror("accept");
        /* FIXME: Now what? */
        sleep(1);
        if(++RetryCount<50)goto RetryAccept;
    }
    
    Type          = SockTypeConnected;
    Peer.Protocol = SockProtocolTCP;
    fd            = sock;
    
    DataPacket pak;
    pak.Type     = DataPacket::PacketTypeOpen;
    pak.Peer     = Peer;
    return pak;
}

DataPacket Socket::ReceiveUDP()
{
    Datatype Buffer[16384];
    DataPacket pak;
    
    Peer.AddrLen = sizeof(Peer.Addr);

    int Size = recvfrom(fd, Buffer, sizeof(Buffer), 0, (struct sockaddr *)&Peer.Addr, &Peer.AddrLen);
    if(Size < 0)
    {
        perror("recvfrom");
        /* FIXME: What now? */
        return pak;
    }
    
    pak.Type     = DataPacket::PacketTypeData;
    pak.Peer     = Peer;
    pak.MakeData(Buffer, Size);
    return pak;
}

DataPacket Socket::ReceiveTCP()
{
    Datatype Buffer[16384];
    DataPacket pak;
    
    int Size = read(fd, Buffer, sizeof(Buffer));
    if(Size < 0)
    {
        perror("read");
        /* FIXME: What now? */
        return pak;
    }
    if(!Size)
    {
        /* Empty packet (possible EOF). Ignore it. */
        return pak;
    }
    
    pak.Type     = DataPacket::PacketTypeData;
    pak.Peer     = Peer;
    pak.MakeData(Buffer, Size);
    return pak;
}

DataPacket Socket::Die()
{
    close(fd);
    
    DataPacket pak;
    
    if(Type == SockTypeConnected)
    {
        pak.Type     = DataPacket::PacketTypeClose;
        pak.Peer     = Peer;
    }
    
    return pak;
}

void Socket::SendTCP(const vector<Datatype> &data)
{
    SendQueue.Feed(&data[0], data.size());
}

void Socket::SendUDP(const vector<Datatype> &data)
{
    /* FIXME TODO: Peer is only temporary!
     * This data must be sent *immediately*!
     * Or then we should save the Peer somehow.
     */
    SendQueue.Feed(&data[0], data.size());
}

void Socket::SendTCP()
{
    const Datatype *buf = SendQueue.GetBegin();
    unsigned        len = SendQueue.GetSize();
    
    int status = write(fd, buf, len);
    if(status < 0)
    {
        /* FIXME: What should we do instead? */
        perror("write");
        return;
    }
    
    if(status > 0)
        SendQueue.Eat(status);
}

void Socket::SendUDP()
{
    const Datatype *buf = SendQueue.GetBegin();
    unsigned        len = SendQueue.GetSize();
    
    int status = sendto(fd, buf, len, 0, (const struct sockaddr *)&Peer.Addr, Peer.AddrLen);
    if(status < 0)
    {
        /* FIXME: What should we do instead? */
        perror("sendto");
        return;
    }
    
    if(status > 0)
        SendQueue.Eat(status);
}

int Socket::CheckError()
{
    int code;
    unsigned len = sizeof(code);
    getsockopt(fd, SOL_SOCKET, SO_ERROR, &code, &len);
    errno = code;
    
    switch(Type)
    {
        case SockTypeConnecting:
            perror("connect");
            break;
        case SockTypeConnected:
            break;
        case SockTypeListening:
            break;
    }
    return code;
}
